
/* Tetris v0.4 - Tetris game for VGA PCs running Linux
 * Copyright (C) 1995 Adam P. Jenkins
 *
 * Board.cc: methods for the Board class, implementing tetris board.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <vga.h>
#include <vgagl.h>
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>

#include "Board.h"

#ifdef DEBUG
extern ofstream debugf;
#endif

Board::Board(int x, int y, int w, int h, int sqw, int sqh, int bg, char *title)
{
#ifdef DEBUG
   debugf.form("%s called\n", __PRETTY_FUNCTION__);
   debugf.form("x = %d, y = %d, w = %d, h = %d, sqw = %d, sqh = %d, bg = %d\n",
                x, y, w, h, sqw, sqh, bg);  
   debugf.flush();
#endif
   // leave 4 extra rows at top so pieces can start offscreen
   h += 4;
   Squares = new Square* [h]; 
   for (int i = 0; i < h; i++)
      Squares[i] = new Square[w];
   Square::setBg(bg);
   Square::setWidth(sqw);
   Square::setHeight(sqh);

   X = x;
   Y = y;
   Width = w * sqw + 2 * BorderWidth;
   Height = (h - 4) * sqh + 2 * BorderWidth;

   BoardWidth = w;
   BoardHeight = h;
   Touched = true;

   if (title)
   {
      Title = new char[strlen(title) + 1];
      strcpy(Title, title);
   }
}

Board::~Board()
{
#ifdef DEBUG
   debugf.form("%s called\n", __PRETTY_FUNCTION__);
#endif
   for (int i = 0; i < BoardHeight; i++)
      delete [] Squares[i];

   delete [] Squares;
   delete [] Title;
}

// places a piece on the board, but doesn't update the screen.  Doesn't check
// if the piece fits, because canPlace() should have been called first.
void Board::placePiece(const Piece& p, int orient, int x, int y, bool falling)
{
#ifdef DEBUG
   debugf.form("%s called with orient = %d, x = %d, y = %d\n",
                __PRETTY_FUNCTION__, orient, x, y);
   debugf.flush();
#endif
   int boardx, boardy;
   y += 4;
   
   if (fits(p, orient, x, y) == false)
      return;

   for (int i = 0; i < NO_SQUARES; i++) 
   {
      boardx = x + p.index[orient][i][0];
      boardy = y + p.index[orient][i][1];
      Squares[boardy][boardx].setActive(true);
      Squares[boardy][boardx].setColor(p.color);
      Squares[boardy][boardx].setFalling(falling);
   }
}

// erases a piece from the screen
void Board::erasePiece(const Piece& p, int orient, int x, int y)
{
#ifdef DEBUG
   debugf.form("%s called with orient = %d, x = %d, y = %d\n",
                __PRETTY_FUNCTION__, orient, x, y);
   debugf.flush();
#endif
   int boardx, boardy;
   y += 4;

   if (fits(p, orient, x, y) == false)
      return;
   
   for (int i = 0; i < NO_SQUARES; i++) 
   {
      boardx = x + p.index[orient][i][0];
      boardy = y + p.index[orient][i][1];
      Squares[boardy][boardx].setActive(false);
   }
}

void Board::eraseRow(int y)
{
#ifdef DEBUG
   debugf.form("%s called with y = %d", __PRETTY_FUNCTION__, y);
   debugf.flush();
#endif
   int x, i;

   if (y < 0 || y >= BoardHeight)
      return;

   // to erase a row, simply move all the rows above it, down one
   for (i = y; i > 0; i--) 
      for (x = 0; x < BoardWidth; x++) 
	 Squares[i][x] = Squares[i-1][x];

   for (x = 0; x < BoardWidth; x++) // take care of the top row
      Squares[0][x].setActive(false);
}

// erases all full rows, and returns the number of rows erased
int Board::eraseRows()
{
#ifdef DEBUG
   debugf.form("%s called\n", __PRETTY_FUNCTION__);
   debugf.flush();
#endif
   bool full;
   int num = 0;

   for (int y = BoardHeight - 1; y >= 0; y--) 
   {
      full = true;
      for (int x = 0; x < BoardWidth; x++) 
	 if (Squares[y][x].active() == false) 
	 {
	    full = false;
	    break;
	 }
      if (full) 
      {
	 eraseRow(y);
	 y++;			// because all the rows have moved down
	 num++;
      }
   }
   return num;
}

// returns true if the piece fits on the board (doesn't check if it overlaps
// another piece).
bool Board::fits(const Piece& p, int orient, int x, int y)
{
#ifdef DEBUG
   debugf.form("%s called with orient = %d, x = %d, y = %d\n",
                __PRETTY_FUNCTION__, orient, x, y);
   debugf.flush();
#endif
   int boardx, boardy;
   
   for (int i = 0; i < NO_SQUARES; i++) 
   {
      boardx = x + p.index[orient][i][0];
      boardy = y + p.index[orient][i][1];
      if (boardx < 0 || boardx >= BoardWidth ||
	  boardy < 0 || boardy >= BoardHeight)
	 return false;
   }
   return true;
}
   
bool Board::canPlace(const Piece& p, int orient, int x, int y)
{
#ifdef DEBUG
   debugf.form("%s called with orient = %d, x = %d, y = %d\n",
                __PRETTY_FUNCTION__, orient, x, y);
   debugf.flush();
#endif
   int boardx, boardy;
   y += 4;

   if (fits(p, orient, x, y) == false) // first check if it fits on board
      return false;
   
   for (int i = 0; i < NO_SQUARES; i++)  // see if it overlaps another piece
   {
      boardx = x + p.index[orient][i][0];
      boardy = y + p.index[orient][i][1];
      if (Squares[boardy][boardx].active() 
          && !Squares[boardy][boardx].falling())
	 return false;
   }
   
   return true;
}  

void Board::update()
{
#ifdef DEBUG
   debugf.form("%s called\n", __PRETTY_FUNCTION__);
   debugf.flush();
#endif
   int screenx, screeny;

   if (Touched)
   {
      gl_fillbox(X, Y, Width, Height, 0);
      DrawBorder(Title);
      Touched = false;
   }
   
   for (int y = 4; y < BoardHeight; y++) 
      for (int x = 0; x < BoardWidth; x++) 
	 if (Squares[y][x].changed()) 
	 {
	    screenx = X + BorderWidth + (x * Square::width());
	    screeny = Y + BorderWidth + ((y - 4) * Square::height());
	    Squares[y][x].display(screenx, screeny);
	 }
}

void Board::touch()
{
#ifdef DEBUG
   debugf.form("%s called\n", __PRETTY_FUNCTION__);
   debugf.flush();
#endif

   Touched = true;

   for (int y = 4; y < BoardHeight; y++)
      for (int x = 0; x < BoardWidth; x++)
	 Squares[y][x].touch();
}

void Board::clear()
{
#ifdef DEBUG
   debugf.form("%s called\n", __PRETTY_FUNCTION__);
   debugf.flush();
#endif
   for (int y = 0; y < BoardHeight; y++)
      for (int x = 0; x < BoardWidth; x++)
	 Squares[y][x].setActive(false);
}

