
/* Tetris v0.4 - Tetris game for VGA PCs running Linux
 * Copyright (C) 1995 Adam P. Jenkins
 *
 * Game.cc: File containing main game loop, and supporting functions.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <Pix.h>

#include <Keyboard.h>
#include <Button.h>
#include <Mouse.h>
#include <Timer.h>
#include <MsgBox.h>

#include "Board.h"
#include "Pieces.h"
#include "tetris.h"

// return a random int between 0 and n
inline int randomn(int n)
{
   return random() % n;
}

// updates the MainBoard and BoardShadow on screen.
void updateBoard()
{
   int x = Mouse::x();
   int y = Mouse::y();
   if (Mouse::ok() && 
       x > MainBoard->x() && x < MainBoard->x() + MainBoard->width()
       && y > MainBoard->y() && y < BoardShadow->y() + BoardShadow->height()) 
   {
      Mouse::setVisible(false);
      MainBoard->update();
      if (game_stats.BoardShadow())
         BoardShadow->update();
      Mouse::setVisible(true);
   } 
   else
   {
      MainBoard->update();
      if (game_stats.BoardShadow())
         BoardShadow->update();
   }
}

// updates BoardShadow, but doesn't redraw it on screen.
inline void updateShadow(const Piece& p, int orient, int x)
{
   BoardShadow->clear();

   // fill in all x coords from piece 
   for (int i = 0; i < NO_SQUARES; i++)
      BoardShadow->setSquare(p.index[orient][i][0] + x, 0, p.color);
}

inline void PauseLoop()
{
   MsgBox m(true, "GAME PAUSED\n\n(Click here or press Enter to resume)");
   m.activate();
}

// randomly fill in squares in 'height' rows, for GAME2
void fillSomeSquares(int height)
{
   int x, y, color;
   
   for (int i = 0; i < height * (MainBoard->boardwidth()/2); i++) 
   {
      x = randomn(MainBoard->boardwidth());
      y = (MainBoard->boardheight() - randomn(height)) - 1;
      color = Pieces[randomn(NO_PIECES)].color;
      MainBoard->setSquare(x, y, color);
   }
}

inline int GameSpeed(int level)
{
   double l = level;
   return (int)(32-l*(3-l/12));
}

// level is the level to start at, type is the type of game to play
void playGame(int level)
{
   int x, y;			// current x/y of falling piece
   int orient;			// current orient of falling piece
   int nextorient;		// orient of next piece
   int piece_no;		// piece number of falling piece
   int nextpiece_no;		// piece number of next piece
   int cycles_left;		// number of cycles between moving piece
   bool dropping = false;	// true if piece is being dropped
   int levels_dropped;		// no. levels piece has been dropped
   bool quit = false;
   bool done;

   if (setGameKeyboard()) 
   {
      cerr << "Failed to initialize raw keyboard\n";
      return;
   }

   MainBoard->clear(); 		// clear board
   if (game_stats.Type2())
      fillSomeSquares(game_stats.height());
 
   MainBoard->update();

   // reset game stats
   game_stats.setState(GameInfo::play);
   game_stats.setPoints(0);
   game_stats.setLevel(level);
   if (game_stats.Type2())
      game_stats.setLines(25);
   else
      game_stats.setLines(0);
   updateStatus();  

   setHelp();			// set help box to game commands
   HelpBox->update();
  
   Timer::init();		// start timer to control framerate

   // get coords for first piece
   nextorient = 1;
   nextpiece_no = randomn(NO_PIECES);

   // this is the main game loop.  It ends when pieces reach the top of the
   // board, or the quit command is given
   do {
      x = MainBoard->boardwidth()>>1; // start piece in middle
      y = -2;			 // start it falling off screen

      orient = nextorient;
      piece_no = nextpiece_no;
      
      // calculate what the next piece will be in advance, so can display 
      // it in next piece display
      nextorient = 1;
      nextpiece_no = randomn(NO_PIECES);
      
      cycles_left = GameSpeed(game_stats.level());
      levels_dropped = 0;
      done = false;	// done with this piece

      // display next piece and shadow
      if (Mouse::ok())
         Mouse::setVisible(false);
      NextPiece->clear();
      NextPiece->placePiece(Pieces[nextpiece_no], nextorient, 2, 1);
      if (game_stats.NextPiece())
	 NextPiece->update();
      updateShadow(Pieces[piece_no], orient, x);
      if (Mouse::ok())
         Mouse::setVisible(true);

      if (!MainBoard->canPlace(Pieces[piece_no], orient, x, y)) 
      {
	 quit = true;
	 done = true;
      } 
      else 
      {
	 MainBoard->placePiece(Pieces[piece_no], orient, x, y);
	 updateBoard();
      }

      // this loop continues until a piece reaches the bottom.  It handles
      // commands and dropping the piece.
      while (!done) 		// while not done with piece
      {
	 Keyboard::update();
	 // move piece left
	 if (Keyboard::keyhit(SCANCODE_CURSORLEFT)) 
	 {
	    if (MainBoard->canPlace(Pieces[piece_no], orient, x - 1, y)) 
	    {
	       MainBoard->erasePiece(Pieces[piece_no], orient, x, y);
	       x -= 1;
	       MainBoard->placePiece(Pieces[piece_no], orient, x, y);
	       updateShadow(Pieces[piece_no], orient, x);
	    }
	 // move piece right
	 } 
	 else if (Keyboard::keyhit(SCANCODE_CURSORRIGHT)) 
	 {
	    if (MainBoard->canPlace(Pieces[piece_no], orient, x + 1, y)) 
	    {
	       MainBoard->erasePiece(Pieces[piece_no], orient, x, y);
	       x += 1;
	       MainBoard->placePiece(Pieces[piece_no], orient, x, y);
	       updateShadow(Pieces[piece_no], orient, x);
	    }
	 // rotate counter-clockwise
	 } 
	 else if (Keyboard::keyhit(SCANCODE_CURSORUP)) 
	 {
	    if (MainBoard->canPlace(Pieces[piece_no], (orient+1)%4, x, y)) 
	    {
	       MainBoard->erasePiece(Pieces[piece_no], orient, x, y);
	       orient = (orient + 1) % 4;
	       MainBoard->placePiece(Pieces[piece_no], orient, x, y);
	       updateShadow(Pieces[piece_no], orient, x);
	    }
	 // rotate clockwise
	 } 
	 else if (Keyboard::keyhit(SCANCODE_CURSORDOWN)) 
	 {
	    if (MainBoard->canPlace(Pieces[piece_no], (orient+3)%4, x, y)) 
	    {
	       MainBoard->erasePiece(Pieces[piece_no], orient, x, y);
	       orient = (orient + 3) % 4;
	       MainBoard->placePiece(Pieces[piece_no], orient, x, y);
	       updateShadow(Pieces[piece_no], orient, x);
	    }
	 } 
	 else if (Keyboard::keyhit(SCANCODE('p'))) 
	 {
	    PauseLoop();
	 } 
	 else if (Keyboard::keyhit(SCANCODE_Q)) 
	 {
	    quit = true;
	    done = true;
	 } 
	 else if (Keyboard::keyhit(SCANCODE_SPACE)) 
	 {
	    if (y >= 0) 
	    {
	       cycles_left = 0;   // make it fall a square immediately
	       dropping = true;
	    }
	 }
	 else if (Keyboard::keyhit(SCANCODE('n')))
	 {
	    toggleNextPiece();
	 }
	 else if (Keyboard::keyhit(SCANCODE('b')))
	 {
	    toggleBoardShadow();
	 }

         if (Mouse::ok() && Mouse::update() && Mouse::leftb()) 
         {
            for (Pix i = ButtonList.first(); i != 0; ButtonList.next(i))
               if (ButtonList(i)->inside(Mouse::x(), Mouse::y()))
               {
                  switch (ButtonList(i)->id()) {
                  case PAUSE_CMD:
                     PauseLoop();
                     break;
                  case QUIT_CMD:
		     quit = true;
		     done = true;
		     break;
		  } // switch
		  
		  break;
               } 
	 } // if (Mouse::ok())

	 updateBoard();	// update screen

         if (dropping) 
         {
            levels_dropped += 1;
            dropping = false;
         } 
         else 
         {
            levels_dropped = 0;
         }
	 cycles_left -= 1;
	 if (cycles_left <= 0) 
	 {
	    cycles_left = GameSpeed(game_stats.level());
	    if (MainBoard->canPlace(Pieces[piece_no], orient, x, y + 1)) 
	    {
	       MainBoard->erasePiece(Pieces[piece_no], orient, x, y);
	       y += 1;
	       MainBoard->placePiece(Pieces[piece_no], orient, x, y);
	    } 
	    else   // if piece is no longer falling, set falling to false
	    {
	       MainBoard->placePiece(Pieces[piece_no], orient, x, y, false);
	       done = true;
	    }
	 }

	 updateBoard();
	 Timer::sync();
      }				// while (!done) with piece

      game_stats.incPoints(levels_dropped);
      {
         int tmp;
         tmp = MainBoard->eraseRows();  // returns number of rows erased
         if (tmp > 0) 
         {
            switch (tmp) {
            case 1:
               game_stats.incPoints((game_stats.level() + 1) * 40);
               break;
            case 2:
               game_stats.incPoints((game_stats.level() + 1) * 100);
               break;
            case 3:
               game_stats.incPoints((game_stats.level() + 1) * 300);
               break;
            case 4:
               game_stats.incPoints((game_stats.level() + 1) * 1200);
               break;
            }
            
	    if (game_stats.Type1()) 
	    {
	       game_stats.incLines(tmp);
	       if ((game_stats.lines() / 10) > game_stats.level())
		  game_stats.incLevel();
	    } 
	    else if (game_stats.Type2())
	       game_stats.decLines(tmp);
	    else	// Type 3 game, just increment lines
	       game_stats.incLines(tmp);
         } // if (tmp > 0)...
      }

      // update screen displays
      updateBoard();
      updateStatus();


      if (game_stats.Type2() && game_stats.lines() == 0)
         quit = true;
   } while (!quit);

   if (game_stats.Type2() && game_stats.lines() == 0)
   {
      MsgBox m(true, "YOU WON!");
      m.activate();
   }
   else
   {
      MsgBox m(true, "GAME OVER");
      m.activate();
   }

   Keyboard::restore();
   game_stats.setState(GameInfo::wait);
   setHelp();
   HelpBox->update();
}
