
/* Tetris v0.4 - Tetris game for VGA PCs running Linux
 * Copyright (C) 1995 Adam P. Jenkins
 *
 * DIBImageLoader.cc: class members for DIBImageLoader.  For loading a windows
                      BMP or DIB image.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
// Written by Adam Jenkins
//
// loadFile method for loading a Windows BMP (DIB) file.
//
// File: DIBImageLoader.cc
#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <iostream.h>
#include <fstream.h>
#include "DIBImageLoader.h"

////////////////////////////////////////////////////////////////////////////
// Error strings
char *DIBImageLoader::errlist[7] = {
   "No error", 				// 0
   "corrupt file error",		// 1
   "file open error",			// 2
   "memory allocation error",		// 3
   "read error",			// 4
   "unsuported  picture format",	// 5
   "compressed bitmap"			// 6
   };


////////////////////////////////////////////////////////////////////////////
// Call this after loadHeader or load for testing purposes.
void DIBImageLoader::displayHeader()
{
   cout << "Header info for file " << FileName << endl;
   cout << "--------------------------------------\n";
   cout << "BITMAPFILEINFO {\n";
   cout << "  short bfType = 0x" << hex << bfh.bfType << dec << endl;
   cout << "  int   bfSize = " << bfh.bfSize << endl;
   cout << "  short bfReserved1 = " << bfh.bfReserved1 << endl;
   cout << "  short bfReserved2 = " << bfh.bfReserved2 << endl;
   cout << "  int   bfOffBits = " << bfh.bfOffBits << endl;
   cout << "};\n\n";

   cout << "BITMAPINFOHEADER {\n";
   cout << "  unsigned int   biSize = " << bih.biSize << endl;
   cout << "  unsigned int   biWidth = " << bih.biWidth << endl;
   cout << "  unsigned int   biHeight = " << bih.biHeight << endl;
   cout << "  unsigned short biPlanes = " << bih.biPlanes << endl;
   cout << "  unsigned short biBitCount = " << bih.biBitCount << endl;
   cout << "  unsigned int   biCompression = " << bih.biCompression << endl;
   cout << "  unsigned int   biSizeImage = " << bih.biSizeImage << endl;
   cout << "  unsigned int   biXPelsPerMeter = " << bih.biXPelsPerMeter << endl;
   cout << "  unsigned int   biYPelsPerMeter = " << bih.biYPelsPerMeter << endl;
   cout << "  unsigned int   biClrUsed = " << bih.biClrUsed << endl;
   cout << "  unsigned int   biClrImportant = " << bih.biClrImportant << endl;
   cout << "}\n";
}  // displayHeader

// load BFH and BIH headers from a bmp file.  returns -1 on error, or 0
// on success.
int DIBImageLoader::loadHeader(const char *file)
{
   if (verbose) 
      cerr << __PRETTY_FUNCTION__ << " called with " << file << endl;
   
   if (file == NULL) 
   {
      Verrno = _PICERR_FILEOPEN;
      return -1;
   }
   
   ifstream infile(file, ios::in|ios::nocreate);
   if (!infile) 
   {
      Verrno = _PICERR_FILEOPEN;
      return -1;
   }

   infile.read((char *) &bfh, BFH_SIZE);
   if (infile.gcount() < BFH_SIZE) 
   {
      Verrno = _PICERR_FILEREAD;
      infile.close();
      return -1;
   }

   infile.read(&bih, BIH_SIZE);
   if (infile.gcount() < BIH_SIZE) 
   {
      Verrno = _PICERR_FILEREAD;
      infile.close();
      return -1;
   }

   if (FileName) 
      delete [] FileName;
   FileName = new char [strlen(file) + 1];
   strcpy(FileName, file);

   infile.close();
   Verrno = _PICERR_NOERROR;
   return 0;
}


int DIBImageLoader::loadFile(const char *file, char flags)
{
   int w, h, bpp;  // width and height

   if (verbose) 
      cerr.form("%s called with %s, %d\n", __PRETTY_FUNCTION__, file, flags);

   if (file == NULL) 
   {
      Verrno = _PICERR_FILEOPEN;
      return -1;
   }

   // loadHeader will initialize bfh and bih
   if (loadHeader(file) < 0)	// loadHeader sets Verrno itself
   {
      return -1;
   }

   if (!(flags & (LOAD_IMAGE|LOAD_PALETTE)))  // with flags 0, just load headers
   {
      Verrno = _PICERR_NOERROR;
      return EXIT_SUCCESS;
   }

   if (bfh.bfType != 0x4d42)   // ascii BM
   {
      Verrno = _PICERR_CORRUPT;
      return -1;
   }

   // Run some checks to see if file may be corrupt
   struct stat statbuf;
   if (stat(file, &statbuf) < 0) 
   {
      Verrno = _PICERR_FILEREAD;
      return -1;
   }

   if (statbuf.st_size != bfh.bfSize)
   {
      Verrno = _PICERR_CORRUPT;
      return -1;
   }

   if (bih.biCompression != 0)		// can't handle compressed bitmaps
   {
      Verrno = _PICERR_ISRLE;
      return -1;
   }

   w = bih.biWidth;
   h = bih.biHeight;
   bpp = bih.biBitCount;  // bits per pixel
   if (w == 0 || h == 0) 
   {
      Verrno = _PICERR_CORRUPT;
      return -1;
   }

   if (bpp > 8) 
   {
      Verrno = _PICERR_BADMAGIC;  // can't do more than 256 colors yet
      return -1;
   }
   VnumColors = 1 << bpp;

   ifstream infile(file, ios::in|ios::nocreate);
   if (!infile) 
   {
      Verrno = _PICERR_FILEOPEN;
      return -1;
   }

   // load the palette
   if (flags & (LOAD_PALETTE|USE_RGBPALETTE)) 
   {
      if (loadPalette(infile, flags) < 0)   // loadPalette sets Verrno on error
      {
         errCleanUp();
         infile.close();
         return -1;
      }
   }

   // load the image
   if (flags & LOAD_IMAGE) 
   {
      if (loadImage(infile, flags) < 0) 
      {
	 errCleanUp();
	 infile.close();
	 return -1;
      }
   }       

   if (flags & USE_RGBPALETTE)  // just use palette to get rgb values
      freePalette();
   
   infile.close();
   Verrno = _PICERR_NOERROR;
   return 0;
}


int DIBImageLoader::loadPalette(ifstream& infile, char flags)
{
   if (verbose) 
      cerr << __PRETTY_FUNCTION__ << ": flags = "  << (int)flags << endl;

   if (!infile.seekg((streamoff)(BFH_SIZE + BIH_SIZE), ios::beg)) 
   {
      Verrno = _PICERR_CORRUPT;
      return -1;
   }

   CPalette *pal = new CPalette(VnumColors);
   char dummy;
   if (pal == NULL) 
   {
      Verrno = _PICERR_NOMEM;
      return -1;
   }
   
   for (int i = 0; i < VnumColors; i++) 
   {
      u_char clr[3];
      infile.read(clr, 3);
      infile.get(dummy);
      if (!infile) 
      {
         delete pal;
         Verrno = _PICERR_FILEREAD;
         return -1;
      }
      // colors are stored in the file BGR, but we want them RGB
      if (flags & USE_RGBPALETTE) 
         pal->setColor(i, clr[2], clr[1], clr[0]);
      else
         pal->setColor(i, clr[2] >> 2, clr[1] >> 2, clr[0] >> 2);
   }

   Vpalette = pal;
   VpaletteByteSize = VnumColors * 3;

   return 0;
}

int DIBImageLoader::loadImage(ifstream& infile, char flags)
{
   if (verbose) 
      cerr << __PRETTY_FUNCTION__ << ": flags = " << (int)flags << endl;

   if (!infile.seekg((streamoff)bfh.bfOffBits, ios::beg)) 
   {
      Verrno = _PICERR_CORRUPT;
      return -1;
   }

   unsigned char *bmap = new unsigned char [bih.biWidth * bih.biHeight];
   if (bmap == NULL) 
   {
      Verrno = _PICERR_NOMEM;
      return -1;
   }

   int x, y, count, w, h = bih.biHeight;
   unsigned char *ptr = bmap, in, r, g, b;
   
   // for all formats, 'w' is set to the actual width of each row in the file,
   // as opposed to the width of the picture, bih.biWidth.  This is because
   // each row is aligned on a 32 bit boundary in the file, so there may be
   // extra bytes at the end of each row.
   //
   // Rows are stored bottom row first in the file, though each row is stored
   // leftmost pixel first.
   switch (bih.biBitCount) {
   case 1:       // 2 color bitmap
      w = 8 * (bfh.bfSize - bfh.bfOffBits) / h;
      for (y = h - 1; y >= 0; y--) 
      {
	 for (x = 0, count = 0; x < (w / 8); x++) 
	 {
	    if (!infile.get(in)) 
	    {
	       delete [] bmap;
	       Verrno = _PICERR_FILEREAD;
	       return -1;
	    }

            // each byte represents eight pixels, so now read each bit,
            // most significant bit first.
	    for (int k = 7; k >= 0; k--) 
	    {
	       if (count < bih.biWidth) 
	       {
		  if (flags&USE_RGBPALETTE) 
		  {
		     Vpalette->getRGB(1 & (in >> k), r, g, b);
		     ptr[count + y * bih.biWidth] =
			CPalette::getColor(r, g, b);
		  } 
		  else
		     ptr[count + y * bih.biWidth] = 1 & (in >> k);
	       }
	       count++;
	    } // for (k...
	 } // for (x...
      } // for (y...

      break;
      
   case 4:    // 16 color bitmap
      w = 2 * (bfh.bfSize - bfh.bfOffBits) / h;
      for (y = h - 1; y >= 0; y--) 
      {
	 for (x = 0, count = 0; x < (w / 2); x++) 
	 {
	    if (!infile.get(in)) 
	    {
	       Verrno = _PICERR_FILEREAD;
	       delete [] bmap;
	       return -1;
	    }

            // read the most significant 4 bits
	    if (count < bih.biWidth) 
	    {
	       if (flags&USE_RGBPALETTE) 
	       {
		  Vpalette->getRGB(in >> 4, r, g, b);
		  ptr[count + y * bih.biWidth] = CPalette::getColor(r, g, b);
	       } 
	       else
		  ptr[count + y * bih.biWidth] = in >> 4;
	    }
	    count++;

	    // read the least significant 4 bits
	    if (count < bih.biWidth) 
	    {
	       if (flags&USE_RGBPALETTE) 
	       {
		  Vpalette->getRGB(in & 0x0f, r, g, b);
		  ptr[count + y * bih.biWidth] = CPalette::getColor(r, g, b);
	       } 
	       else
		  ptr[count + y * bih.biWidth] = in & 0x0f;
	    }
	    count++;
	 }  // for (x...
      }  // for (y...

      break;

   case 8:     // 256 color bitmap
      w = (bfh.bfSize - bfh.bfOffBits) / h;     
      for (y = h - 1; y >= 0; y--) {
	 for (x = 0, count = 0; x < w; x++) 
	 {
	    if (!infile.get(in)) 
	    {
	       Verrno = _PICERR_FILEREAD;
	       delete [] bmap;
	       return -1;
	    }
	    
	    if (count < bih.biWidth) 
	    {
	       if (flags&USE_RGBPALETTE) 
	       {
		  Vpalette->getRGB(in, r, g, b);
		  ptr[count + y * bih.biWidth] = CPalette::getColor(r, g, b);
	       } 
	       else
		  ptr[count + y * bih.biWidth] = in;
	    }
	    count++;
	 }  // for (x...
      } // for (y...
      
      break;
   }  // switch
   
   Vimage = bmap;
   Vwidth = bih.biWidth;
   Vheight = bih.biHeight;
   VimageByteSize = bih.biWidth * bih.biHeight;

   return 0;
}

