// -*-c++-*-

/* Tetris v0.4 - Tetris game for VGA PCs running Linux
 * Copyright (C) 1995 Adam P. Jenkins
 *
 * DIBImageLoader.h: Class derived from ImageLoader, to load DIB (windows
                     bitmaps) into a byte-per-pixel bitmap.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
// Written by Adam Jenkins.
//
// Class derived from ImageLoader in Bitmap.h to load Windows bmp (DIB) 
// format files.
//
// File:  DIBImageLoader.h

#ifndef __DIBImageLoader__
#define __DIBImageLoader__

#include <stdlib.h>
#include <stdio.h>
#include <fstream.h>
#include <sys/types.h>
#include "CPalette.h"
#include "Bitmap.h"

// define error codes to be returned be loadFile
#define _PICERR_NOERROR 0
#define _PICERR_CORRUPT 1
#define _PICERR_FILEOPEN 2
#define _PICERR_NOMEM 3
#define _PICERR_FILEREAD 4
#define _PICERR_BADMAGIC 5
#define _PICERR_ISRLE 6

// extra flag for load
#define USE_RGBPALETTE 4

// Headers at the beginning of a bmp file.
#define BFH_SIZE 14   
// making these packed makes it easier to read the struct from the file.
// though it may not be portable code.
typedef struct BITMAPFILEHEADER {
   short bfType __attribute__ ((packed));  // Type of file, must be BM (0x4d42)     
   int   bfSize __attribute__ ((packed));  // Size of file, in bytes    
   short bfReserved1 __attribute__ ((packed));  // Reserved, should be zero
   short bfReserved2 __attribute__ ((packed));  // same
   // byte offset from beginning of file to actual image data, 
   // (skips palette)
   int   bfOffBits __attribute__ ((packed));    
} BITMAPFILEHEADER;

#define BIH_SIZE 40
typedef struct BITMAPINFOHEADER{
   unsigned int  biSize;          // size of BIH structure
   unsigned int  biWidth;         // width of bitmap in pixels 
   unsigned int  biHeight;        // height of bitmap in pixels
   unsigned short   biPlanes;     // number of planes ??, must be 1
   unsigned short   biBitCount;   // bits per pixel, 1, 4, 8, or 24
   unsigned int  biCompression;   // whether it's compressed or not
   unsigned int  biSizeImage;     // size of image in bytes, usually 0
   unsigned int  biXPelsPerMeter; // target pixels per meter, X
   unsigned int  biYPelsPerMeter; // ditto, Y
   unsigned int  biClrUsed;       // num colors actually used, or 0 if all
   unsigned int  biClrImportant;  // 0 if all colors are important
} BITMAPINFOHEADER;


// Most of the methods for this class are defined in ImageLoader, in 
// Bitmap.h.  Use load(char *file, flags) to load a file.  flags can be
// one or more of LOAD_IMAGE, LOAD_PALETTE, and USE_RGBPALETTE.  
// USE_RGBPALETTE  will cause the bitmap colors to be set from a standard 
// RGB palette, which only approximates the colors in the palette.
class DIBImageLoader : public ImageLoader
{
 public:
   // load is declared in ImageLoader
   
   DIBImageLoader()
   {
      FileName = NULL;
      Verrno = 0;
      verbose = false;
   }
   ~DIBImageLoader()
     {
	freeImage();
	freePalette();
	if (FileName) delete [] FileName;
     }
   int      loadHeader(const char *file);
   void     displayHeader();
   void     setVerbose(bool v) { verbose = v; }
   char *   ImageName() const { return FileName; }
   char *   perror() const { return errlist[Verrno]; }

 private:
   // loadFile is called by load to actually load the file.
   int      loadFile(const char *file, char flags);
   int      loadPalette(ifstream& infile, char flags);
   int      loadImage(ifstream& infile, char flags);
   void     errCleanUp(void) {    // called to clean up after an error
      freeImage();
      freePalette();
      delete [] FileName;
      FileName = NULL;
   }
   char *   FileName;      // image file name
   static char *errlist[7];   
   int Verrno; 
   bool verbose;

   
   BITMAPFILEHEADER bfh;   // File info header.
   BITMAPINFOHEADER bih;   // picture info header.
};

#endif   // __DIBImageLoader__
