
/* Tetris v0.4 - Tetris game for VGA PCs running Linux
 * Copyright (C) 1995 Adam P. Jenkins
 *
 * DisplayBox.cc: DisplayBox member functions.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
// Methods for the displaybox class

#include <vga.h>
#include <vgagl.h>
#include <stdio.h>
#include <string.h>
#include <Pix.h>

#include "Ui.h"
#include "Colors.h"
#include "DisplayBox.h"


DisplayBox::~DisplayBox()
{
   delete [] Title;
   delete [] UnderScreen;
   while (!TextLines.empty())
   {
      delete TextLines.front();
      TextLines.del_front();
   }
}

// breaks up 'text' into lines.  If 'text' contains newlines, it will be
// split on the newlines, otherwise it will be split in lines which are at
// most (Width - 2 * BorderWidth) long, on word boundaries; which also means
// that Width must already have been set
// Returns: the length of the longest line
int DisplayBox::WrapText(const char *const text)
{
   if (!text)
      return -1;
   
   int maxlen = 0;		// to record length of longest line
   String buf = text;
   String *s;

   while (!TextLines.empty())
   {
      delete TextLines.front();
      TextLines.del_front();
   }
   

   if (buf.contains('\n')) 	// break up string on newlines
   {
      do {
	 s = new String(buf.before("\n"));
	 TextLines.append(s);
	 buf = buf.after("\n");
	 maxlen = maxlen >? s->length();
      } while (buf.contains('\n'));
      
      s = new String(buf);	// take care of last line
      TextLines.append(s);
      maxlen = maxlen >? s->length();
   }
   else if (Width > 0)		// break it up on word boundaries
   {
      const int textwidth = (Width - 2 * BorderWidth) / Ui::fontWidth();
      while (buf.length() > textwidth)
      {
         // finds the word-boundary (whitespace) closest to textwidth
	 int idx = buf.index(RXwhite, -(buf.length() - textwidth));
	 if (idx <= 0)		// can't break on a word boundary
	 {
	    s = new String(buf.before(textwidth));
	    buf.del(0, textwidth);
	 }
	 else		// break on word boundary
	 {
	    s = new String(buf.before(idx));
	    buf = buf.after(RXwhite, idx);
	 }
	 TextLines.append(s);
	 maxlen = maxlen >? s->length();
      }				// while
      
      if (buf.length() > 0) 	// take care of last line
      {
	 s = new String(buf);
	 TextLines.append(s);
	 maxlen = maxlen >? s->length();
      }
   }
   else				// no newlines and Width hasn't been set
   {
      s = new String(buf);
      TextLines.append(s);
      maxlen = s->length();
   }
   
   return maxlen;
}

// Prints the contents of TextLines, beginning at 'y1', inside the box.
// If 'centered' is true, centers each line in the box.  If any lines are too
// long to fit in the box, they are truncated
void DisplayBox::PrintTextLines(int y1, int y2, bool centered) const
{
   if (Width == 0)
      return;

   int screeny = y1;
   int screenx;

   if (TextLines.empty())
      return;

   gl_setwritemode(WRITEMODE_MASKED);

   // the max number of chars a line can be
   const int maxlen = (Width - 2 * BorderWidth) / Ui::fontWidth();
   
   for (Pix i = TextLines.first(); i != 0; TextLines.next(i))
   {
      // center the line if centered is true, and line isn't too long
      if (centered && TextLines(i)->length() <= maxlen)
         screenx = X + (Width - TextLines(i)->length() * Ui::fontWidth()) / 2;
      else
         screenx = X + BorderWidth;

      // stop printing lines if the line would go out of vertical bounds
      if (screeny + Ui::fontHeight() > y2)
	 break;

      // print line. If line is too long, only print maxlen chars
      if (TextLines(i)->length() > maxlen)
	 gl_writen(screenx, screeny, maxlen, (char *)TextLines(i)->chars());
      else
	 gl_write(screenx, screeny, (char *)TextLines(i)->chars());

      screeny += Ui::fontHeight(); // move screeny to next line
   }  // for
}



void DisplayBox::DrawBorder(const char *title) const
{
   if (Width == 0 || Height == 0)
      return;
   
   // draw top line, with title if there's a title
   if (title)
   {
      gl_setwritemode(WRITEMODE_MASKED);
      int titlewidth = strlen(title) * Ui::fontWidth();
      int titlex = X + (Width - titlewidth)/2;
      gl_hline(X, Y,     titlex - 4, Clr::WHITE);
      gl_hline(X, Y + 1, titlex - 4, Clr::WHITE);
      
      gl_write(titlex, (Y - Ui::fontHeight()) + 6, (char *)title);
      
      gl_hline(titlex + titlewidth + 3, Y,     X + Width - 1, Clr::WHITE);
      gl_hline(titlex + titlewidth + 3, Y + 1, X + Width - 1, Clr::WHITE);
   }
   else
   {
      gl_hline(X, Y,     X + Width - 1, Clr::WHITE);
      gl_hline(X, Y + 1, X + Width - 1, Clr::WHITE);
   }

   // bottom line
   gl_hline(X, Y + Height - 1, X + Width - 1, Clr::WHITE);
   gl_hline(X, Y + Height - 2, X + Width - 1, Clr::WHITE);

   // left side
   gl_line(X,     Y, X,     Y + Height - 1, Clr::WHITE);
   gl_line(X + 1, Y, X + 1, Y + Height - 1, Clr::WHITE);

   // right side
   gl_line(X + Width - 1, Y, X + Width - 1, Y + Height - 1, Clr::WHITE);
   gl_line(X + Width - 2, Y, X + Width - 2, Y + Height - 1, Clr::WHITE);
}

void DisplayBox::SaveUnder() const
{
   if (UnderScreen)
      delete [] UnderScreen;

   UnderScreen = new char[Width * Height];
   if (UnderScreen)
      gl_getbox(X, Y, Width, Height, UnderScreen);
}

void DisplayBox::RestoreUnder() const
{
   if (UnderScreen)
   {
      gl_putbox(X, Y, Width, Height, UnderScreen);
      delete [] UnderScreen;
      UnderScreen = NULL;
   }
}
