
/* Tetris v0.4 - Tetris game for VGA PCs running Linux
 * Copyright (C) 1995 Adam P. Jenkins
 *
 * Menu.cc: Menu member functions.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <string.h>
#include <stdio.h>
#include <Pix.h>

#include "Mouse.h"
#include "Ui.h"
#include "Colors.h"
#include "Keyboard.h"

#include "Menu.h"

void Menu::init(int x, int y, int width, int height, char *title, char *text)
{
   X = x;
   Y = y;
   Width = width;
   Height = height;
   if (title)
   {
      Title = new char[strlen(title) + 1];
      strcpy(Title, title);
   } 
   
   if (text)
   {
      Text = new char[strlen(text) + 1];
      strcpy(Text, text);

      if (Width > 0)
	 WrapText(Text);
   } 
   else
      Text = NULL;

   CancelBn = new Button(0, 0, "Cancel", 0, Clr::WHITE, Clr::PURPLE);
   Height = Height >? FindHeight();
}

Menu::Menu(int x, int y, int width, int height, char *title, char *text)
{
   Centered = false;
   MinWidth = width;
   MinHeight = height;
   init(x, y, width, height, title, text);
}

Menu::Menu(int x, int y, char *title, char *text)
{
   Centered = false;
   MinWidth = MinHeight = 0;
   init(x, y, 0, 0, title, text);
}

Menu::Menu(int width, char *title, char *text)
{
   Centered = true;
   MinWidth = width;
   MinHeight = 0;
   init(0, 0, width, 0, title, text);
}

// Menu will be centered on screen
Menu::Menu(char *title, char *text)
{
   Centered = true;
   MinWidth = MinHeight = 0;
   init(0, 0, 0, 0, title, text);
}

Menu::~Menu()
{
   if (Title)
      delete [] Title;

   if (Text)
      delete [] Text;

   while (!Items.empty())
   {
      delete [] Items.front().label;
      Items.del_front();
   }
}

int Menu::FindHeight() const
{
   int height;

   // I leave 16 pixels between different parts of the menu, for instance,
   // between the Text at the top and the menu items, and between the menu
   // items and the cancel button
   if (!TextLines.empty())
      height = (TextLines.length() + Items.length()) * Ui::fontHeight()
	 + CancelBn->height() + 2 * BorderWidth + 2 * MenuSepHeight;
   else
      height = Items.length() * Ui::fontHeight() + CancelBn->height()
	 + 2 * BorderWidth + MenuSepHeight;

   return height;
}

int Menu::FindWidth() const
{
   int width = 0;

   for (Pix i = Items.first(); i != 0; Items.next(i))
      width = width >? strlen(Items(i).label);

   return (width + 4) * Ui::fontWidth() + 2 * BorderWidth;
}

int Menu::FindItemsY() const
{
   int itemsy;

   if (!TextLines.empty())
      itemsy = Y + BorderWidth 
             + TextLines.length() * Ui::fontHeight() + MenuSepHeight;
   else
      itemsy = Y + BorderWidth;

   return itemsy;
}

void Menu::add(char *label, int key, int id, MenuClbkFn fn, void *data)
{
   if (!label)
      return;
   
   char *tmp = new char[strlen(label) + 1];
   strcpy(tmp, label);

   MenuItem item = {tmp, key, id, menu_item, fn, data, NULL};
   add(item);
}

void Menu::add(char *label, int key, int id, Menu *submenu)
{
   if (!label)
      return;

   char *tmp = new char[strlen(label) + 1];
   strcpy(tmp, label);

   MenuItem item = {tmp, key, id, sub_menu, NULL, NULL, submenu};
   add(item);
}

void Menu::add(MenuItem& item)
{
   Items.append(item);

   int width = FindWidth();
   if (width > Width)
   {
      Width = width;
      if (Text)
	 WrapText(Text);
   }

   Height = Height >? MinHeight >? FindHeight();

   if (Centered)
   {
      X = (WIDTH - Width) / 2;
      Y = (HEIGHT - Height) / 2;
   }
   
   ItemsY = FindItemsY();
}

void Menu::remove(int id)
{
   for (Pix i = Items.first(); i != 0; Items.next(i))
   {
      if (Items(i).id == id)
      {
	 remove(i);
	 break;
      }
   }
}

void Menu::remove(char *label)
{
   for (Pix i = Items.first(); i != 0; Items.next(i))
   {
      if (!strcmp(Items(i).label, label))
      {
	 remove(i);
	 break;
      }
   }
}

void Menu::remove(Pix& i)
{
   delete [] Items(i).label;
   Items.del(i);

   // recalculate width and height
   int width = FindWidth();

   if (width < Width && width > MinWidth)
   {
      Width = width;
      if (Text)
	 WrapText(Text);
   }

   Height = FindHeight();

   if (Centered)
   {
      X = (WIDTH - Width) / 2;
      Y = (HEIGHT - Height) / 2;
   }
   
   ItemsY = FindItemsY();
}

void Menu::printItem(int num, const MenuItem& item, bool selected) const
{
   if (!item.label || num < 0 || num >= Items.length())
      return;

   int x = X + BorderWidth;
   int y = ItemsY + num * Ui::fontHeight();
   
   gl_fillbox(x, y, Width - 2 * BorderWidth, Ui::fontHeight(),
	      selected ? Clr::PURPLE : Clr::GRAY);

   gl_write(x, y, (char *)item.label);
   gl_writen(X + (Width - (BorderWidth + 3 * Ui::fontWidth() + 1)), y, 
             1, (char *)&item.key);
             
   if (item.type == sub_menu)
      gl_write(X + (Width - (BorderWidth + Ui::fontWidth() + 1)), y, ">");
}
   
   
int Menu::display() const
{
   Pix i;
   int item;

   gl_fillbox(X, Y, Width, Height, Clr::GRAY);
   
   // Draw a border for the menu
   DrawBorder();
  
   gl_setwritemode(WRITEMODE_MASKED);
   
   if (!TextLines.empty())
   {
      int texty = Y + BorderWidth;
      int textheight = TextLines.length() * Ui::fontHeight();
      PrintTextLines(texty, texty + textheight, false);
      gl_hline(X, texty + textheight + MenuSepHeight/2, X + Width - 1, Clr::WHITE);
   }

   gl_hline(X, ItemsY + Items.length()*Ui::fontHeight() + MenuSepHeight/2,
	    X + Width - 1, Clr::WHITE);
   CancelBn->display(X + ((Width - CancelBn->width()) / 2),
		     (Y + Height) - (CancelBn->height() + BorderWidth));
   
   if (Items.empty())
      return -1;
   
   // print first item as selected
   i = Items.first();
   printItem(0, Items(i), true);
   
   // print the rest unselected
   for (Items.next(i), item = 1; i != 0; Items.next(i), item++) {
      printItem(item, Items(i), false);
   }

   return 0;
}

int Menu::activate() const
{
   if (Mouse::ok())
      Mouse::setVisible(false);
   SaveUnder();			// save screen under menu
   display();
   if (Mouse::ok())
      Mouse::setVisible(true);
      

   Pix indices[Items.length()], i;
   int current = 0, j;

   for (i = Items.first(), j = 0; i != 0; Items.next(i), j++)
      indices[j] = i;

   int result = -1;			// will be set to menu choice
   bool quit = false;
   while (!quit)
   {
      if (Keyboard::update())   // parse keyboard
      {
	 if (Keyboard::keypressed(SCANCODE_CURSORBLOCKUP) ||
	     Keyboard::keypressed(SCANCODE_CURSORUP)) 
	 {       // Move selected item up one
	    if (!Items.empty() && current > 0)
	    {
	       if (Mouse::ok())
		  Mouse::setVisible(false);
	       printItem(current, Items(indices[current]), false);
	       current -= 1;
	       printItem(current, Items(indices[current]), true);
	       if (Mouse::ok())
		  Mouse::setVisible(true);
	    }
	 }
	 else if (Keyboard::keypressed(SCANCODE_CURSORBLOCKDOWN) ||
	          Keyboard::keypressed(SCANCODE_CURSORDOWN))
	 {       // Move selected item down one.
	    if (!Items.empty() && current < Items.length() - 1)
	    {
	       if (Mouse::ok())
		  Mouse::setVisible(false);
	       printItem(current, Items(indices[current]), false);
	       current += 1;
	       printItem(current, Items(indices[current]), true);
	       if (Mouse::ok())
		  Mouse::setVisible(true);
	    }
	 }
	 else if (Keyboard::keypressed(SCANCODE_ESCAPE))
	 {
	    quit = true;
	    result = -1;
	 }
	 else if (Keyboard::keypressed(SCANCODE_ENTER))
	 {    // return item currently selected
	    quit = true;
	    if (Items.empty())
	       result = -1;
	    else
	       result = Items(indices[current]).id;
	 }
	 else
	 {			// scan for hotkeys
	    for (i = Items.first(); i != 0; Items.next(i))
	    {
	       if (Keyboard::keypressed( SCANCODE(Items(i).key) ))
	       {
		  quit = true;
		  result = Items(i).id;
	       }
	    }
	 }
      }  // if (Keyboard::update()....

      if (Mouse::ok() && Mouse::update()) {
	 int x = Mouse::x();
	 int y = Mouse::y();

	 // check if on a menu item, move highlight
	 
	 if (!Items.empty() && x > X && x < (X + Width - 1) && y > ItemsY &&
	     y < (ItemsY + Items.length() * Ui::fontHeight())) 
	 {
	    int item = (y - ItemsY) / Ui::fontHeight();
	    if (item != current)  // move highlight
	    {
	       Mouse::setVisible(false);
	       printItem(current, Items(indices[current]), false);
	       current = item;
	       printItem(current, Items(indices[current]), true);
	       Mouse::setVisible(true);
	    }
	    if (Mouse::leftb())
	    {
	       quit = true;
	       result = Items(indices[current]).id;
	    }
	 }
	 else if (Mouse::leftb() && CancelBn->inside(x, y))
	 {
	    quit = true;
	    result = -1;
	 }
      }  //if (Mouse::ok() ....
   }  // while (!quit) ...

   if (result >= 0)		// call the callback function, if not NULL
   {
      i = indices[current];
      if (Items(i).type == menu_item && Items(i).fn != NULL)
         Items(i).fn(result, Items(i).data);
      else if (Items(i).type == sub_menu && Items(i).submenu != NULL)
         Items(i).submenu->activate();
   }
   
   if (Mouse::ok())
      Mouse::setVisible(false);
   RestoreUnder();		// restore screen under menu
   if (Mouse::ok())
      Mouse::setVisible(true);

   return result;
}

