// -*-c++-*-

/* Tetris v0.4 - Tetris game for VGA PCs running Linux
 * Copyright (C) 1995 Adam P. Jenkins
 *
 * Menu.h: Menu class definition. Derived from DisplayBox.  Displays a menu,
 *         User can select items.  Supports callback functions, submenus, or
 *         can just return a menu item id.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
// A menu class which uses the mouse or keyboard to select items.
// begun on: Sun Aug 13 00:36:10 EDT 1995

#ifndef __Menu__
#define __Menu__

#include <stdio.h>
#include <DLList.h>
#include <Pix.h>

#include "Button.h"
#include "DisplayBox.h"

// Menu callback function type
typedef void (*MenuClbkFn)(int, void *);

class Menu : public DisplayBox {
public:
   // width and height are MINIMUMs.  If items are added which don't fit,
   // width and height will be adjusted as necessary, but if items are removed,
   // width and height won't go below the parameters below.
   Menu(int x, int y, int width, int height, char *title, char *text = NULL);
   
   // width and height will be set according to item lengths
   Menu(int x, int y, char *title, char *text = NULL);

   // width won't go below width
   Menu(int width, char *title, char *text = NULL);

   // x and y will be set so the menu appears centered on the screen.  width
   // and height will be set automatically
   Menu(char *title, char *text = NULL);
   ~Menu();

   // add a menu item.  label is what appears on the menu.  key is a hot key
   // to activate the menu item while the menu is active. id is the value
   // returned if this item is selected. Id must be >= 0, because less than 0
   // is used to return error codes, and using an id < 0 would mess up this 
   // scheme.
   // fn is a pointer to a callback funtion.  If not null, it will be called
   // when this menu item is selected, and passed the value of id.  see typedef
   // above. data can be a pointer to any arbitrary data that should also be 
   // passed to the callback function.
   void add(char *label, int key, int id, MenuClbkFn fn = NULL, 
            void *data = NULL);
   // Like the above, except instead of a callback function, pass a pointer
   // to another menu to be called if this item is selected.
   void add(char *label, int key, int id, Menu *submenu);

   void remove(int id);		  // remove item by id
   void remove(char *label);	  // remove item by name

   // Displays and runs the menu.  Returns the id of the item selected,
   // or -1 if ESC was pressed, or the Cancel button was clicked.
   int activate() const;

private:
   const int MenuSepHeight = 16; // height between sections of menu
   // variables
   int MinWidth;		// minimum width and height
   int MinHeight;
   int ItemsY;			// y coord of beginning of items

   char *Text;			// Text which appears at top of menu
   bool Centered;		// if true, center menu on screen
   
   Button *CancelBn;
   
   enum ItemType {
      menu_item,
      sub_menu
   };
   
   struct MenuItem {
      char *label;		// item label
      int key;			// hot key
      int id;			// return value of this item
      ItemType type;		// item or sub_menu
      MenuClbkFn fn;		// callback function for this item
      void *data;		// arbitrary data passed to callback fn
      Menu *submenu;		// pointer to a submenu
   };

   DLList<MenuItem> Items;

   // functions
   void init(int x, int y, int width, int height, char *title, char *text);

   void add(MenuItem& item);

   // prints Menu item number 'num', prints it as selected if 'selected' is
   // true.  Uses 'num' to determine the y coord to print at.
   void printItem(int num, const MenuItem& item, bool selected) const;

   // This function actually removes a menu item
   void remove(Pix& i);

   // displays the menu, returns 0 if successful, -1 on error or if there
   // are no menu items
   int display() const; 

   // calculates the height of the menu based on TextLines and Items.
   int FindHeight() const;

   // calculate width of Menu base on Menu item length
   int FindWidth() const;

   // calculate new ItemsY
   int FindItemsY() const;
};
   
   
#endif  // __Menu__
