
/* Tetris v0.4 - Tetris game for VGA PCs running Linux
 * Copyright (C) 1995 Adam P. Jenkins
 *
 * tetris.cc: Contains main(), and initialization functions.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
// My first attempt at a video game.

#include <vga.h>
#include <vgagl.h>
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>  // for toupper
#include <getopt.h>
#include <SLList.h>
#include <Pix.h>

// files included from libgame
#include <Ui.h>
#include <Button.h>
#include <Mouse.h>
#include <Keyboard.h>
#include <Timer.h>
#include <HiScore.h>
#include <Colors.h>
#include <InfoBox.h>
#include <Menu.h>
#include <SBitmap.h>

#include "Board.h"
#include "Game.h"
#include "tetris.h"
#include "utils.h"
#include "config.h"


// Globals
SLList<Button*> ButtonList;
Board *MainBoard, *NextPiece, *BoardShadow;
InfoBox *StatusBox, *HelpBox;
GameInfo game_stats;

#ifdef DEBUG
ofstream debugf("tetris.log");
#endif

// sets the contents of the HelpBox for different game modes
void setHelp()
{
   static char mainhelp[] = 
"S    start game
Q    quit
M    game menu
N    toggle next piece display
B    toggle board shadow";
   
   static char playhelp[] = 
"\030     rotate counter-clockwise
\031     rotate clockwise
\033     move left
\032     move right
SPACE drop piece
P     pause game
Q     quit game
N     toggle next piece display
B     toggle board shadow"; 

   static char pausehelp[] = 
"R    resume game"; 

   if (game_stats.pauseState()) 
      HelpBox->setContents(pausehelp);
   else if (game_stats.waitState()) 
      HelpBox->setContents(mainhelp);
   else if (game_stats.playState())
      HelpBox->setContents(playhelp);
}

// updates the contents of StatusBox based on info from game_stats
void updateStatus()
{
   char buf[128];
   int type;
  
   if (game_stats.Type1())
      type = 1;
   else if (game_stats.Type2())
      type = 2;
   else
      type = 3;

   sprintf(buf,
"Score:  
  %06u

Lines:
  %03d

Level:
  %02d

Game Type:
  Game %d", 
           game_stats.points(), game_stats.lines(), game_stats.level(), type);
   StatusBox->setContents(buf);
   StatusBox->update();
}

// Sets up raw keyboard mode and key delays for the main game loop
int setGameKeyboard()
{
   if (Keyboard::init() < 0)
      return -1;
      
   Keyboard::setdelay(SCANCODE_CURSORUP, FPS/10);
   Keyboard::setinitdelay(SCANCODE_CURSORUP, FPS/2);
   
   Keyboard::setdelay(SCANCODE_CURSORLEFT, FPS/20);
   Keyboard::setinitdelay(SCANCODE_CURSORLEFT, FPS/4);
   
   Keyboard::setdelay(SCANCODE_CURSORRIGHT, FPS/20);
   Keyboard::setinitdelay(SCANCODE_CURSORRIGHT, FPS/4);
   
   Keyboard::setdelay(SCANCODE_CURSORDOWN, FPS/10);
   Keyboard::setinitdelay(SCANCODE_CURSORDOWN, FPS/2);
   
   Keyboard::setdelay(SCANCODE_Q, 40);
   Keyboard::setdelay(SCANCODE('n'), 40);
   Keyboard::setdelay(SCANCODE('b'), 40);
   Keyboard::setdelay(SCANCODE('p'), 40);

   keyboard_translatekeys(TRANSLATE_CURSORKEYS);

   return 0;
}

// Redraws the screen.  All objects on screen 'know' what they should look
// like.
void drawScreen()
{
#ifdef DEBUG
   debugf.form("Entered %s\n", __PRETTY_FUNCTION__);
   debugf.flush();
#endif
   if (Mouse::ok())
      Mouse::setVisible(false);

   gl_clearscreen(BG_COLOR);  // purple
   
   // print title
   char *title = "Tetris";
   PrintMag(WIDTH/2 - strlen(title)*Ui::fontWidth(), 10, 2, title);

   MainBoard->touch();
   MainBoard->update();
   
   if (game_stats.NextPiece())
   {
      NextPiece->touch();
      NextPiece->update();
   }

   if (game_stats.BoardShadow())
   {
      BoardShadow->touch();
      BoardShadow->update();
   }

#ifdef DEBUG
   debugf << "Updated Boards\n" << flush;
#endif

   StatusBox->display();
   updateStatus();
   HelpBox->display();
#ifdef DEBUG
   debugf << "Updated InfoBoxes\n" << flush;
#endif

   for (Pix i = ButtonList.first(); i != 0; ButtonList.next(i))
      ButtonList(i)->display();
#ifdef DEBUG
   debugf << "Displayed Buttons\n" << flush;
#endif

   if (Mouse::ok())
      Mouse::setVisible(true);
#ifdef DEBUG
   debugf << "Made mouse cursor visible\n" << flush;
#endif
} 

void toggleNextPiece()
{
   if (game_stats.NextPiece())
   {
      gl_fillbox(NextPiece->x(), NextPiece->y() - Ui::fontHeight(), 
                 NextPiece->width(), NextPiece->height() + Ui::fontHeight(), 
		 BG_COLOR);
      game_stats.toggleNextPiece();
   }
   else
   {
      NextPiece->touch();
      NextPiece->update();
      game_stats.toggleNextPiece();
   }
}

void toggleBoardShadow()
{
   if (game_stats.BoardShadow())
   {
      gl_fillbox(BoardShadow->x(), BoardShadow->y(), BoardShadow->width(),
                 BoardShadow->height(), BG_COLOR);
      game_stats.toggleBoardShadow();
   }
   else
   {
      BoardShadow->touch();
      BoardShadow->update();
      game_stats.toggleBoardShadow();
   }
}

///////// Callback functions for Menus

// callback fn for Level Menu; sets level
void LevelMenuClbk(int id, void *data)
{
   if (game_stats.setLevel(id) >= 0)
   {
      *(int *)data = id;	// data points to the level var, local to main
      updateStatus();
   }
}

// callback fn for game type menu
void TypeMenuClbk(int id, void *data)
{
   if (game_stats.setType(id) >= 0)
      updateStatus();
}

// callback fn for height menu; sets height of random blocks for game 2.
void HeightMenuClbk(int id, void *data)
{
   // This menu only gets called through the Type Menu, so it sets game type.
   game_stats.setType(GameInfo::game2);
   updateStatus();
   game_stats.setHeight(id);
}

// callback fn for High Scores menu item.
// Displays the highscores screen, and redraws screen when done.
void showScoresClbk(int id, void *data)
{
  Keyboard::restore();
  showScores(*(HiScore *)data);
  Keyboard::init();
  drawScreen();
}
      
// adds items to Level menu.
void makeLevelMenu(Menu& m, int *level)
{
   m.add("Level zero",   '0', 0, LevelMenuClbk, level);
   m.add("Level one",    '1', 1, LevelMenuClbk, level);
   m.add("Level two",    '2', 2, LevelMenuClbk, level);
   m.add("Level three",  '3', 3, LevelMenuClbk, level);
   m.add("Level four",   '4', 4, LevelMenuClbk, level);
   m.add("Level five",   '5', 5, LevelMenuClbk, level);
   m.add("Level six",    '6', 6, LevelMenuClbk, level);
   m.add("Level seven",  '7', 7, LevelMenuClbk, level);
   m.add("Level eight",  '8', 8, LevelMenuClbk, level);
   m.add("Level nine",   '9', 9, LevelMenuClbk, level);
   m.add("Level ten",    't', 10, LevelMenuClbk, level);
   m.add("Level eleven", 'e', 11, LevelMenuClbk, level);
   m.add("Level twelve", 'w', 12, LevelMenuClbk, level);
   m.add("Level thirteen", 'h', 13, LevelMenuClbk, level);
   m.add("Level fourteen", 'f', 14, LevelMenuClbk, level);
   m.add("Level fifteen", 'i', 15, LevelMenuClbk, level);
   m.add("Level sixteen", 's', 16, LevelMenuClbk, level);
}

// adds height menu items.
void makeHeightMenu(Menu &m)
{
   m.add("One",   '1', 1, HeightMenuClbk);
   m.add("Two",   '2', 2, HeightMenuClbk);
   m.add("Three", '3', 3, HeightMenuClbk);
   m.add("Four",  '4', 4, HeightMenuClbk);
   m.add("Five",  '5', 5, HeightMenuClbk);
   m.add("Six",   '6', 6, HeightMenuClbk);
   m.add("Seven", '7', 7, HeightMenuClbk);
   m.add("Eight", '8', 8, HeightMenuClbk);
   m.add("Nine",  '9', 9, HeightMenuClbk);
}
   
// Called only once to initialize svgalib and create objects.
void initGame()
{
   Ui::init(G640x480x256);
   Ui::drawToPhysical();
#ifdef DEBUG
   debugf.form("%s: initialized screen\n", __PRETTY_FUNCTION__);
   debugf.flush();
#endif
   if (loadFont(FONTFILE, FONTWIDTH, FONTHEIGHT)) 
   {
      cerr << "Failed to load font file\n";
      CleanUp();
      exit(1);
   }
#ifdef DEBUG
   debugf << "Loaded font\n" << flush;
#endif

   MainBoard = new Board(BOARDX, BOARDY, BOARDWIDTH, BOARDHEIGHT, 
                         SQUAREWIDTH, SQUAREHEIGHT, Clr::GRAY);
   BoardShadow = new Board(BOARDX, BOARDY + MainBoard->height(), BOARDWIDTH, 1,
                      SQUAREWIDTH, SQUAREHEIGHT, Clr::GRAY);
   NextPiece = new Board(BOARDX + MainBoard->width() + 32, BOARDY, 6, 5,
                         SQUAREWIDTH, SQUAREHEIGHT, Clr::GRAY, "Next Piece");
#ifdef DEBUG
   debugf << "Created Boards\n" << flush;
#endif



   Button *Bptr;
   // Make a list of the buttons.  The order here is important
   Bptr = new Button(BUTTONX, BOARDY, BUTTONWIDTH, "Quit", QUIT_CMD);
   ButtonList.append(Bptr);
#ifdef DEBUG
   debugf << "Created Quit Button, Bptr = " << Bptr << endl << flush;
#endif
   
   Bptr = new Button(BUTTONX, Bptr->y2() + 16, BUTTONWIDTH, "Pause", PAUSE_CMD);
   ButtonList.append(Bptr);
#ifdef DEBUG
   debugf << "Created Pause Button, Bptr = " << Bptr << endl << flush;
#endif

   Bptr = new Button(BUTTONX, Bptr->y2() + 16, BUTTONWIDTH, "New Game", PLAY_CMD);
   ButtonList.append(Bptr);
#ifdef DEBUG
   debugf << "Created New Game Button, Bptr = " << Bptr << endl << flush;
   debugf << "New Game Button y2 = " << Bptr->y2() << endl << flush;
#endif

   Bptr = new Button(BUTTONX, Bptr->y2() + 16, BUTTONWIDTH, "Game Menu", MENU_CMD);
   ButtonList.append(Bptr);
#ifdef DEBUG
   debugf << "Created Game Menu Button, Bptr = " << Bptr << endl << flush;
   
   debugf << "Created buttons and button list\n" << flush;
#endif

   StatusBox = new InfoBox(STATUSX, Bptr->y2() + 32, 
                           STATUSWIDTH, STATUSHEIGHT, false, "STATUS");
   HelpBox = new InfoBox(MainBoard->x() + MainBoard->width() + 24,
                         NextPiece->y() + NextPiece->height() + 24,
                         HELPWIDTH, HELPHEIGHT, false, "KEYBOARD COMMANDS");

   Mouse::init(MakeMouseCursor(Clr::BLACK, Clr::WHITE), MOUSE_DEV);	
   if (Mouse::ok()) 
   {
      Mouse::setposition(WIDTH/2, BOARDY/2);
   }
#ifdef DEBUG
   debugf << "Initialized Mouse and created mouse pointer\n" << flush;
#endif

   srandom(time(NULL));    // seed random generator
   Timer::setFrameRate(FPS);
   setHelp();
#ifdef DEBUG
   debugf << "Set help and timer\n" << flush;
#endif


}

// deletes all objects, restores text mode, and restores keyboard mode.
void CleanUp()
{
   // delete Boards
   delete MainBoard;
   delete NextPiece;
   delete BoardShadow;

   // delete Buttons
   while (!ButtonList.empty()) 
   {
      delete ButtonList.front();
      ButtonList.del_front();
   }
   
   if (Mouse::ok())
      Mouse::close();

   if (Keyboard::initd())
      Keyboard::restore();
      
   Ui::restore();
}
   
int main(int argc, char **argv)
{
   int level = 0;
   int opt, opt_ind;
   struct option options[] = 
   {
      {"level",     1, 0, 'l'},
      {"game_type", 1, 0, 't'},
      {"paula",     0, 0, 'p'}
   };
   
   game_stats.setType(GameInfo::game1);
   game_stats.setHeight(5);
   
   while ((opt = getopt_long(argc, argv, "l:t:p", options, &opt_ind)) != EOF) 
   {
      switch (opt) {
      case 'l':
         if (game_stats.setLevel(atoi(optarg)) < 0) {
            cerr.form("%s: Level must be 0-16\n", argv[0]);
            exit(1);
         }
         level = game_stats.level();
         break;
      case 't':
         if (game_stats.setType(atoi(optarg)) < 0) {
            cerr.form("%s: valid game types are:
            1  -- Normal tetris game
            2  -- Drop some pieces first
            3  -- Don't change levels
            \n", argv[0]);
            exit(1);
         }
         break;
      case 'p':
         game_stats.setType(GameInfo::game3);
         break;
      case '?':
	 cerr << argv[0] << ": Options are -- 
	 -l, --level <level>
	    Start game at <level>, instead of 0.
	 -t, --game_type <type>
	    Play game type <type>.  Game types are:
	    1 -- Normal tetris (the default).
	    2 -- Drop some pieces first.
	    3 -- Like normal, but don't change levels.
	 -p, --paula
	    Paula type game :-)
	 \n";
	 exit(1);
      } // switch
   }  // while

   initGame();

   drawScreen();

   HiScore scores(SCOREDIR"/tetris.scores");

 //////////// Create Menus //////////////////////////

   Menu LevelMenu(150, "Level", "Choose a starting level");
   makeLevelMenu(LevelMenu, &level);
#ifdef DEBUG
   debugf.form("%s: Created LevelMenu\n", __PRETTY_FUNCTION__);
   debugf.flush();
#endif

   Menu HeightMenu(150, "Height", "Choose Height of Random Blocks");
   makeHeightMenu(HeightMenu);

   Menu TypeMenu(150, "Game Type", "Choose a game type");
   TypeMenu.add("Game one: Normal tetris", '1', 1, TypeMenuClbk);
   TypeMenu.add("Game two: Fill some squares", '2', 2, &HeightMenu);
   TypeMenu.add("Game three: Don't increase levels", '3', 3, TypeMenuClbk);
   TypeMenu.add("Paula's Game", 'p', 3, TypeMenuClbk);

#ifdef DEBUG
   debugf << "Created TypeMenu\n" << flush;
#endif

   Menu GameMenu(150, "Game Menu");
   GameMenu.add("Show High Scores", 'h', SCORES_CMD, showScoresClbk, &scores);
   GameMenu.add("Set Play Level", 'l', LEVEL_CMD, &LevelMenu);
   GameMenu.add("Set Game Type", 't', TYPE_CMD, &TypeMenu);
   
   int command;
   bool quit = false;
   
   // Here's the main loop.  It goes until quit is selected
   while (!quit) 
   {
      command = NO_CMD;

      switch (toupper(vga_getkey())) 
      {
      case 'S':
         command = PLAY_CMD;   break;
      case 'Q':
         command = QUIT_CMD;   break;
      case 'M':
         command = MENU_CMD;   break;
      case 'N':		// toggle next piece
         command = NEXTP_CMD;  break;
      case 'B':		// toggle board shadow
         command = BOARDSH_CMD; break;
      } // switch

      if (Mouse::ok() && Mouse::update() && Mouse::leftb()) 
      {
	 for (Pix i = ButtonList.first(); i != 0; ButtonList.next(i))
	    if (ButtonList(i)->inside(Mouse::x(), Mouse::y())) 
	    {
	       command = ButtonList(i)->id();
	       break;
	    }
#ifdef DEBUG
	    debugf << "Checked Buttons\n" << flush;
#endif
      } // Mouse::ok()

      switch (command) {
      case QUIT_CMD:
         quit = true;
         break;
      case PLAY_CMD:
         playGame(level);

         // GAME3 is too easy so don't add it to high score
         if (!game_stats.Type3() && scores.isHigh(game_stats.points()) != -1) 
         {
            scores.add(getHighScoreName(), game_stats.points());
            showScores(scores);
            drawScreen();
         }
         break;
      case MENU_CMD:
         Keyboard::init();
         GameMenu.activate();
         Keyboard::restore();
         break;
      case NEXTP_CMD:
         toggleNextPiece();
	 break;
      case BOARDSH_CMD:
         toggleBoardShadow();
	 break;
      }  // switch
   } // while

   CleanUp();
}
